"""
THIS IS THE CODE THAT WAS DEVELOPED AT NOON

Find the mode of a list: the most repeated value in the list

Complexity (or expected cost) of operations


sort values in a list L:
    N log N
    (more expensive than linear)
    
List L has n values 

value in List: 
Max value in a List
count of a value in a list
    n comparisons in the worst case
    O(N) : linear
    
Set S has n values

value in Set S:
    1 comparison 
    
Dictionary D has n keys

value in D: ##in the keys for D
    1 comparison
    

List solution:  O(n logn)
    Sort the list first
    (same values are consecutive)
    Go through the list once
    and count occurrences of values
    (and keep track of the current
    value with max count)

Set solution (O (kn)) 
    for ech unique value
        count the number of occurrences of
        that value
        keep track of the maximum

Dictionary solution:  O(n+k)
     O(n) to go through the list
     go through each value
     construct a dictionary with
         key: value
         value: count of that value
    

     find the max count in the values
     O(k): there are as many values as
     distinct values in the list
     
     go through the dictionary keys (there
     are k of them) and if their count
     is the max, then return them/one of them  
     O(k)

Design choices:
   - Multiple values with the same
    frequency: return one or all?
    
   - Total number of values: n
     Number of distinct values: k
""" 

import random, time

def list_sol(L):
    """ Overall complexity: O(n log n)
        Returns the last value with the highest frequency

    """

    L.sort()   # O(nlogn)
    
    maxfreq_val = -1  ##mod value so far
    maxfreq_cnt = 0   ##frequency of mod value so far
    
    cur_val = L[0]  ##current value I am couting
    cur_cnt = 0  ##count for this value
    for val in L:  ##O(n)
        if val == cur_val:
            cur_cnt += 1
        else:
            if cur_cnt > maxfreq_cnt:
                maxfreq_cnt = cur_cnt
                maxfreq_val = cur_val
            cur_cnt = 1
            cur_val = val
    if cur_cnt > maxfreq_cnt:
        maxfreq_cnt = cur_cnt
        maxfreq_val = cur_val
    return (maxfreq_val, maxfreq_cnt)

def list_sol_all(L):
    """ Overall complexity: O(n log n)
        Returns all the values with the highest frequency

    """
    
    L.sort()   # O(nlogn)
    
    maxfreq_val = []  ##mod value so far
    maxfreq_cnt = 0   ##frequency of mod value so far
    cur_val = L[0]  ##current value I am couting
    cur_cnt = 0  ##count for this value
    for val in L:  ##O(n)
        if val == cur_val:
            cur_cnt += 1
        else:
            if cur_cnt > maxfreq_cnt:
                maxfreq_cnt = cur_cnt
                maxfreq_val = [cur_val]
            elif cur_cnt == maxfreq_cnt:
                maxfreq_val.append( cur_val )
            cur_cnt = 1
            cur_val = val
    if cur_cnt > maxfreq_cnt:
        maxfreq_cnt = cur_cnt
        maxfreq_val = cur_val
    elif cur_cnt == maxfreq_cnt:
        maxfreq_val.append( cur_val )    
    return (maxfreq_val, maxfreq_cnt)

def set_sol(L):
    """ Overall complexity: O(k*n)
        For high values of k, becomes an expensive solution.
        Returns one value with the highest frequency, it can be any 
        value as sets are unordered.

    """
    
    maxfreq_val = -1  ##mod value so far
    maxfreq_cnt = 0   ##frequency of mod value so far

    for val in set(L):  #k unique values
        cnt = L.count(val)  #count within n value
        if cnt > maxfreq_cnt:
            maxfreq_cnt = cnt
            maxfreq_val = val
    return (maxfreq_val, maxfreq_cnt)


def set_sol_all(L):
    """ Same complexity as set_sol: O(k*n)
        Returns all the values with the highest frequency

    """
    maxfreq_val = []  ##mod value so far
    maxfreq_cnt = 0   ##frequency of mod value so far

    for val in set(L):  #k unique values
        cnt = L.count(val)  #count within n value
        if cnt > maxfreq_cnt:
            maxfreq_cnt = cnt
            maxfreq_val = [val]
        elif cnt == maxfreq_cnt:
            maxfreq_val.append(val)
    return (maxfreq_val, maxfreq_cnt)


def dict_sol(L):
    """ Overall complexity: O(n+k) which is basically O(n)
        Returns one value with the highest frequency, it can be any 
        value as dictionary keys are stored as sets and hence are 
        unordered.

    """
    
    valdict = dict()
    for val in L:
        if val not in valdict:
            valdict[val] = 1
        else:
            valdict[val] += 1
    maxfreq_cnt = max( valdict.values() )
    for val in valdict:
        if valdict[val] == maxfreq_cnt:
            return (val, maxfreq_cnt)
    
def dict_sol_all(L):
    """ Overall complexity: O(n+k) which is basically O(n)
        Returns all values with the highest frequency.

    """

    valdict = dict()
    for val in L:
        if val not in valdict:
            valdict[val] = 1
        else:
            valdict[val] += 1
    maxfreq_cnt = max( valdict.values() )
    vals = []
    for val in valdict:
        if valdict[val] == maxfreq_cnt:
            vals.append(val)
    return (vals, maxfreq_cnt)


def time_func(f, L):
    """ Calls a function f and then prints it running time.

    """
    
    start = time.time()
    val, cnt = f(L)
    end = time.time()
    print()
    print ("Function: {} took {:.4f} seconds".format(f.__name__, end-start))
    print ("Returned:", val, cnt)

if __name__ == "__main__":
    """ Create a random list. """
    n = 1000
    k = 10
    
    L = []
    for i in range(n):
        L.append ( random.randint(1,k) )
    
    time_func(set_sol, L)
    time_func(set_sol_all, L)
    
    time_func(dict_sol, L)
    time_func(dict_sol_all, L)    
    time_func(list_sol, L)
    
    random.shuffle(L)  ## make sure the list is not sorted
    ## as list_sol changes the ordering of the list. 
    time_func(list_sol_all, L)


    
