"""
THIS IS THE CODE THAT WAS DEVELOPED AT 2PM
<INCLUDES ALSO SOME INEFFICIENT SOLUTIONS FOR COMPARISON.>

Complexity: rough idea of how fast/slow my program is
going to be

Complexity primer: Given a container 
(list/set/dictionary) with n elements

sorting a list:
     O(n log n)

if a value in list L:
count of value X in a List
max value in a list
     n comparisons,  O(n) -> linear
     
     
if a value in a set S:
      1 comparison, O(1) -> Constant time
      
if a value is in dictionary D ( val in D):
      (value is in the keys)
      1 comparison, O(1)
      
if value in list(D.keys()):
      O(n) -> don't do that (because the solution
      above is cheaper and does the same thing)

Problem: Find the mode of a list: the value 
that is  repeated most often
Example: L = [1,2,4,1,2,4,1,1]
1 is repeated 4 times, the most: so it is the
mode

Potential solutions:

List solution: Overall complexity: O(nlogn + n)
     Sort the list
     Then, go through the list
     keep track of count of items and
     max count so far

Set solution: Overall complexity: O(n*k)
     find all unique values (using a set)
     count each one and keep track of max

Dictionary solutiom: Overall complexity: O(n+k)
     construct a dictionary with key as value
     and value as count
     find max count
     return value with max count

Design choices:
     n: number of items in the list
     k: number of unique values in the ist
     Return one value or all values that are
     repeated most

"""

import random, time


#############################################################
### These are the inefficient solutions for comparison    ###
### Make sure you understand why they are inefficient.    ###
#############################################################

def list_sol_nsquared(L):
    ## for each value  in list (n of them)
    ## you count: O(n): total cost O(n*n)
    maxfreq_val = -1
    maxfreq_count = 0
    
    for val in L:
        cnt = L.count(val)
        if cnt > maxfreq_count:
            maxfreq_count = cnt
            maxfreq_val = val
    return maxfreq_val, maxfreq_count


def dict_sol_inefficient(L):
    ## O(n*n)
    val_dict = dict()
    for val in L:   
        if val not in list(val_dict.keys()):
            val_dict[val] = 1
        else:
            val_dict[val] += 1
    maxfreq_count = max( val_dict.values() )
    ##finding max: O(k)
    
    ##going through all keys/values, O(k)
    for val in list(val_dict.keys()): ##go through all keys
        if val_dict[val] == maxfreq_count:
            return val, maxfreq_count

#############################################################
### END OF PURPOSEFULLY INEFFICIENT SOLUTIONS             ###
#############################################################

        
def list_sol(L):
    """ Overall complexity: O(n log n)
        Returns the last value with the highest frequency

    """

    L.sort()  ## O(n logn)
    maxfreq_val = -1
    maxfreq_count = 0
    
    cur_val = L[0]
    cur_count = 0
    for val in L:  ##O(n)
        if val == cur_val:
            cur_count += 1
        else:
            if cur_count > maxfreq_count:
                maxfreq_val = cur_val
                maxfreq_count = cur_count
            
            cur_val = val
            cur_count = 1
    if cur_count > maxfreq_count:
        maxfreq_val = cur_val
        maxfreq_count = cur_count
    return (maxfreq_val, maxfreq_count)

def list_sol_all(L):
    """ Overall complexity: O(n log n)
        Returns all the values with the highest frequency

    """

    L.sort()  ## O(n logn)
    maxfreq_val = []
    maxfreq_count = 0
    
    cur_val = L[0]
    cur_count = 0
    for val in L:  ##O(n)
        if val == cur_val:
            cur_count += 1
        else:
            if cur_count > maxfreq_count:
                maxfreq_val = [cur_val]
                maxfreq_count = cur_count
            elif cur_count == maxfreq_count:
                maxfreq_val.append (cur_val)
            
            cur_val = val
            cur_count = 1
    if cur_count > maxfreq_count:
        maxfreq_val = [cur_val]
        maxfreq_count = cur_count
    elif cur_count == maxfreq_count:
        maxfreq_val.append (cur_val)
        
    return (maxfreq_val, maxfreq_count)


def set_sol(L):
    """ Overall complexity: O(k*n)
        For high values of k, becomes an expensive solution.
        Returns one value with the highest frequency, it can be any 
        value as sets are unordered.

    """
    
    ## set(L) -> O(n)
    ## for each unique value (k of them)
    ## you count: O(n): total cost O(k*n)
    maxfreq_val = -1
    maxfreq_count = 0
    
    for val in set(L):
        cnt = L.count(val)
        if cnt > maxfreq_count:
            maxfreq_count = cnt
            maxfreq_val = val
    return maxfreq_val, maxfreq_count

def set_sol_all(L):
    """ Same complexity as set_sol: O(k*n)
        Returns all the values with the highest frequency

    """
    
    ## Returns all values
    maxfreq_val = []
    maxfreq_count = 0
    
    for val in set(L):
        cnt = L.count(val)
        if cnt > maxfreq_count:
            maxfreq_count = cnt
            maxfreq_val = [val]
        elif cnt == maxfreq_count:
            maxfreq_val.append( val )
    return maxfreq_val, maxfreq_count

    
def dict_sol(L):
    """ Overall complexity: O(n+k) which is basically O(n)
        Returns one value with the highest frequency, it can be any 
        value as dictionary keys are stored as sets and hence are 
        unordered.

    """
    ## O(n+k)
    val_dict = dict()
    for val in L:   ## O(n)
        if val not in val_dict:
            val_dict[val] = 1
        else:
            val_dict[val] += 1
    maxfreq_count = max( val_dict.values() )
    ##finding max: O(k)
    
    ##going through all keys/values, O(k)
    for val in val_dict: ##go through all keys
        if val_dict[val] == maxfreq_count:
            return val, maxfreq_count
        
def dict_sol_all(L):
    """ Overall complexity: O(n+k) which is basically O(n)
        Returns all values with the highest frequency.

    """

    val_dict = dict()
    for val in L:
        if val not in val_dict:
            val_dict[val] = 1
        else:
            val_dict[val] += 1
    maxfreq_count = max( val_dict.values() )
    values = []
    for val in val_dict: ##go through all keys
        if val_dict[val] == maxfreq_count:
            values.append(val)
            
    return values, maxfreq_count

def time_func(f, L):
    """ Calls a function f and then prints it running time.

    """
    start = time.time()
    val, cnt = f(L)
    end = time.time()
    print ()
    print ("Function: {} took {:.4f} seconds".format(f.__name__, end-start))
    print ("Returned", val, cnt)

    

if __name__ == "__main__":
    """ Create a random list. """
    
    n = 100000
    k = 20
    
    L = []
    for i in range(n):
        L.append ( random.randint(1,k) )
            
    time_func(set_sol, L)
    time_func(set_sol_all, L)
    
    time_func(dict_sol, L)
    time_func(dict_sol_all, L)
    time_func(dict_sol_inefficient, L)

    ## Takes too long, so commented out.
    ##time_func(list_sol_nsquared, L)
    time_func(list_sol, L)    
    
    random.shuffle(L)  ## make sure the list is not sorted
    ## as list_sol changes the ordering of the list. 
    time_func(list_sol_all, L)    
    
